#!/usr/local/sbin/charm-env python3
#
# Copyright 2016,2019 Canonical Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Define hw-health charm actions."""

import os
import pathlib
import re
import shlex
import subprocess
import sys
from datetime import datetime

from charmhelpers.core import unitdata
from charmhelpers.core.hookenv import action_fail, action_get, action_set, config, log

from hwhealth import tools


IPMI_SEL = "/usr/sbin/ipmi-sel"
CRON_DIR = pathlib.Path("/etc") / "cron.d"
RUN_PARTS_RE = re.compile("^[a-zA-Z0-9_-]+$")  # run-parts(8)
CRON_RE = re.compile(r"^\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+(\S+)\s+(.*)")


def clear_sel():
    """Clear the IPMI system event log.

    Implementation uses 'ipmi-sel --post-clear'
    Clears the SEL log and stores the cleared entries in action output.
    """
    command = [IPMI_SEL, "--post-clear"]
    try:
        output = subprocess.check_output(command)
        log("Action clear-sel completed, sel log cleared: {}".format(output))
        action_set({"message": output.decode("UTF-8")})
    except subprocess.CalledProcessError as e:
        action_fail("Action failed with {}".format(e))


def show_sel():
    """Show the entries in the IPMI system event log.

    By default, this will show all non-nominal events.
    If you specify show-all, it will show all events.
    """
    show_all = action_get("show-all")
    command = [IPMI_SEL, "--output-event-state"]

    date_start = unitdata.kv().get(tools.Ipmi.SEL_DATE_FILTER_START_KEY)
    if date_start:
        command.extend(["--date-range", "{}-now".format(date_start)])

    try:
        header, body = None, None
        output = subprocess.check_output(command).decode("UTF-8")
        lines = output.splitlines()
        if lines:
            header, body = lines[0], lines[1:]
            if not show_all:
                # This is fairly naive, but it may be good enough for now
                body = [line for line in body if "Nominal" not in line]
        if body:
            final_output = "\n".join([header] + body)
            if date_start:
                final_output += "Has been date filtered: {}".format(date_start)
        else:
            final_output = "No matching entries found"
        log("Action show-sel completed:\n{}".format(final_output))
        action_set({"message": final_output})
    except subprocess.CalledProcessError as e:
        action_fail("Action failed with {}".format(e))


def ack_sel():
    """Set the acknowledgement date filter."""
    message = ""
    if "--date-range" in config("ipmi_check_options"):
        message = (
            "WARNING: Will override user specified --date-range "
            "option on ipmi_check_options config\n"
        )

    date_param = action_get("date")
    if date_param:
        try:
            date = datetime.strptime(date_param, "%Y-%m-%d")
        except ValueError:
            action_fail(
                "Invalid date format. Expected format is ISO 8601 "
                "(YYYY-MM-DD), got: {}".format(date_param)
            )
            return
    else:
        date = datetime.utcnow()

    try:
        unitdata.kv().set(
            tools.Ipmi.SEL_DATE_FILTER_START_KEY, date.strftime("%m/%d/%Y")
        )
        unitdata.kv().flush()
        message += "Only SEL entries newer than {} will be considered".format(
            date.strftime("%Y-%m-%d")
        )
        action_set({"message": message})
        tools.Ipmi().install_cronjob()
    except Exception as e:
        action_fail("Action failed with {}".format(e))


def unack_sel():
    """Unset the acknowledgement date filter."""
    try:
        unitdata.kv().unset(tools.Ipmi.SEL_DATE_FILTER_START_KEY)
        unitdata.kv().flush()
        message = "SEL entries from all times will be considered"
        action_set({"message": message})
        tools.Ipmi().install_cronjob()
    except Exception as e:
        action_fail("Action failed with {}".format(e))


def _get_cron_actions():
    crond_files = CRON_DIR.glob("hwhealth_*")
    return {
        path.stem.split("_", 1)[-1]
        for path in crond_files
        if RUN_PARTS_RE.match(path.name)
    }


def list_cron_actions():
    actions = _get_cron_actions()
    message = "Cron Actions: {}".format(",".join(sorted(actions)))
    log("Action list-cron-actions completed\n" + message)
    action_set({"message": message})


def _run_job_file(fname):
    with fname.open() as f:
        job_lines = f.readlines()
        # match jobs according to crontab(5) man
        # blank, space or tab are all ignorable
        jobs = [m.groups() for m in (CRON_RE.match(line) for line in job_lines) if m]
        outs = []
        for who, cmd in jobs:
            sudo_cmd = ["sudo", "-u", who] + shlex.split(cmd)
            try:
                out = subprocess.check_output(sudo_cmd, stderr=subprocess.STDOUT)
            except subprocess.CalledProcessError as e:
                action_fail("{}: {}\n\t{}".format(who, cmd, str(e)))
                out = e.output + e.stderr
            outs.append((sudo_cmd, out))
        return outs


def run_cron_action():
    cron_entries = {entry for entry in action_get("entries").split(",") if entry}
    cron_actions = _get_cron_actions()
    to_run = cron_entries or cron_actions  # if nothing specified, run all
    wont_run = to_run - cron_actions  # any entries not capable of being run?
    to_run -= wont_run

    results = [
        "{}\n\t{}".format(" ".join(cmd), result)
        for each in to_run
        if each in cron_actions
        for cmd, result in _run_job_file(CRON_DIR / "hwhealth_{}".format(each))
    ]
    response = {}
    if results:
        response["message"] = "\n".join(results)

    if wont_run:
        # skipping the won't runs b/c they were not listed as a cron action
        response["skipped"] = "unavailable actions: {}".format(
            ",".join(sorted(wont_run))
        )
    action_set(response)


# A dictionary of all the defined actions to callables (which take
# parsed arguments).
ACTIONS = {
    "clear-sel": clear_sel,
    "show-sel": show_sel,
    "ack-sel": ack_sel,
    "unack-sel": unack_sel,
    "list-cron-actions": list_cron_actions,
    "run-cron-action": run_cron_action,
}


def main(args):
    """Define main subroutine."""
    action_name = os.path.basename(args[0])

    try:
        ACTIONS[action_name]()
    except Exception as e:
        action_fail(str(e))


if __name__ == "__main__":
    sys.exit(main(sys.argv))
