#!/usr/bin/env python3
"""Check perccli output (gathered via cron) for alertable states."""

import argparse
import json
import os
import sys

from nagios_plugin3 import CriticalError, WarnError, try_check

try:
    # shared lib will be under $NAGIOS_PLUGIN_DIR at runtime
    from hw_health_lib import HWCheckArgumentParser
except ImportError:
    # shared lib will be under $CHARM_SOURCE_DIR/files/common during unit tests
    common_libs_dir = os.path.abspath(
        os.path.join(os.path.dirname(__file__), "..", "common")
    )
    if common_libs_dir not in sys.path:
        sys.path.append(common_libs_dir)
    from hw_health_lib import HWCheckArgumentParser

INPUT_FILE = "/var/lib/nagios/perccli.out"
ARGS = argparse.Namespace()


def handle_results(nlines, match, critical, errors, num_ldrive, num_pdrive, policy):
    """Provide results of checks and raise proper nagios alert."""
    if nlines == 0:
        raise WarnError("WARNING: no output from perccli")
    elif not match:
        raise CriticalError("ERROR: error parsing perccli output")
    elif critical:
        if len(errors) > 0:
            msg = ", ".join(
                [
                    "{}({})".format(cnt, vars()[cnt])
                    for cnt in ("failed_ld", "wrg_policy_ld")
                    if vars().get(cnt, 0) > 0
                ]
            )
            msg += "; ".join(errors)
        else:
            msg = "failure caught but no output available"
        raise CriticalError("CRITICAL: {}".format(msg))
    elif len(errors) > 0:
        raise WarnError("WARNING: {}".format("; ".join(errors)))

    else:
        if num_ldrive == 0:
            msg = "OK: no disks configured for RAID"
        else:
            msg = "OK: Optimal, ldrives[{}], pdrives[{}]".format(num_ldrive, num_pdrive)
            if policy:
                msg += ", policy[{}]".format(policy)
        print(msg)


def parse_output(policy=False):  # noqa:C901
    num_pdrive = num_ldrive = failed_ld = wrg_policy_ld = 0
    nlines = 0
    errors = []
    match = critical = False
    adapter_id = ldrive_id = None

    with open(ARGS.input_file) as devices_json:
        try:
            vd_data = json.load(devices_json)
            match = True
            controllers = vd_data.get("Controllers")
            nlines = len(controllers)
            for controller in controllers:
                resp_data = controller.get("Response Data")
                if resp_data is None:
                    continue
                vdrives = resp_data.get("VD LIST")
                num_pdrive += int(resp_data.get("Physical Drives", 0))
                num_ldrive += int(resp_data.get("Virtual Drives", 0))
                adapter_id = controller.get("Command Status", {}).get("Controller")
                for vd in vdrives:
                    state = vd.get("State")
                    dg, ldrive_id = vd.get("DG/VD").split("/")
                    if state != "Optl":
                        failed_ld += 1
                        msg = "adapter({}):ld({}):state({})".format(
                            adapter_id, ldrive_id, state
                        )
                        errors.append(msg)
                        critical = True
                    if policy:
                        cache_policy = vd.get("Cache")
                        if cache_policy != policy:
                            wrg_policy_ld += 1
                            msg = "adp({}):ld({}):policy({})".format(
                                adapter_id, ldrive_id, cache_policy
                            )
                            errors.append(msg)
                            critical = True
        except json.JSONDecodeError:
            match = False
    handle_results(nlines, match, critical, errors, num_ldrive, num_pdrive, policy)


def parse_args(argv=None):
    """Process CLI arguments."""
    parser = HWCheckArgumentParser(prog="check_perccli", def_input_file=INPUT_FILE)
    return parser.parse_args(args=argv, namespace=ARGS)


def main(argv, policy=False):
    """Define main subroutine."""
    parse_args(argv)
    try_check(parse_output, policy)


if __name__ == "__main__":
    main(sys.argv[1:])
