# Copyright 2019 Canonical Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pathlib
import subprocess
import sys
import tempfile
import unittest
import unittest.mock as mock

sys.path.append(".")
from actions.actions import (  # noqa:E402
    clear_sel,
    list_cron_actions,
    _run_job_file,
    run_cron_action,
    show_sel,
)


class ClearSelTestCase(unittest.TestCase):
    @mock.patch("actions.actions.log")
    @mock.patch("subprocess.check_output")
    @mock.patch("subprocess.check_call")
    def test_clear_sel(self, mock_check_call, mock_subprocess, mock_log):
        sel_output = "Unittest system event log output".encode()
        mock_subprocess.return_value = sel_output
        mock_check_call.return_value = None
        clear_sel()
        mock_check_call.assert_called_once_with(
            ["action-set", "message={}".format(sel_output.decode())]
        )


class ShowSelTestCase(unittest.TestCase):
    @mock.patch("actions.actions.log")
    @mock.patch("actions.actions.action_set")
    @mock.patch("actions.actions.action_get")
    @mock.patch("subprocess.check_output")
    def test_empty_output_from_ipmi_sel(
        self, mock_check_output, mock_action_get, mock_action_set, mock_log
    ):
        show_all_flag = False
        output_body = ""
        expected_output = "No matching entries found"
        self._test_valid_show_sel_call(
            show_all_flag,
            output_body,
            expected_output,
            mock_check_output,
            mock_action_get,
            mock_action_set,
        )

    @mock.patch("actions.actions.log")
    @mock.patch("actions.actions.action_set")
    @mock.patch("actions.actions.action_get")
    @mock.patch("subprocess.check_output")
    def test_only_nominal_entries_with_show_all_false(
        self, mock_check_output, mock_action_get, mock_action_set, mock_log
    ):
        show_all_flag = False
        output_body = "\n".join(
            [
                "Header line",
                "Nominal body line #1",
                "Nominal body line #2",
                "Nominal body line #3",
            ]
        )
        expected_output = "No matching entries found"
        self._test_valid_show_sel_call(
            show_all_flag,
            output_body,
            expected_output,
            mock_check_output,
            mock_action_get,
            mock_action_set,
        )

    @mock.patch("actions.actions.log")
    @mock.patch("actions.actions.action_set")
    @mock.patch("actions.actions.action_get")
    @mock.patch("subprocess.check_output")
    def test_only_nominal_entries_with_show_all_true(
        self, mock_check_output, mock_action_get, mock_action_set, mock_log
    ):
        show_all_flag = True
        output_body = "\n".join(
            [
                "Header line",
                "Nominal body line #1",
                "Nominal body line #2",
                "Nominal body line #3",
            ]
        )
        expected_output = output_body
        self._test_valid_show_sel_call(
            show_all_flag,
            output_body,
            expected_output,
            mock_check_output,
            mock_action_get,
            mock_action_set,
        )

    @mock.patch("actions.actions.log")
    @mock.patch("actions.actions.action_set")
    @mock.patch("actions.actions.action_get")
    @mock.patch("subprocess.check_output")
    def test_non_nominal_entries_present_with_show_all_false(
        self, mock_check_output, mock_action_get, mock_action_set, mock_log
    ):
        show_all_flag = False
        output_body = "\n".join(
            [
                "Header line",
                "Nominal body line #1",
                "Warning line #1",
                "Critical line #1",
                "Nominal body line #2",
                "Nominal body line #3",
                "Warning line #2",
            ]
        )
        expected_output = "\n".join(
            ["Header line", "Warning line #1", "Critical line #1", "Warning line #2"]
        )
        self._test_valid_show_sel_call(
            show_all_flag,
            output_body,
            expected_output,
            mock_check_output,
            mock_action_get,
            mock_action_set,
        )

    def _test_valid_show_sel_call(
        self,
        show_all_flag,
        output_body,
        expected_output,
        mock_check_output,
        mock_action_get,
        mock_action_set,
    ):
        mock_action_get.return_value = show_all_flag
        mock_check_output.return_value = output_body.encode()
        show_sel()
        self.assertEqual(mock_action_set.call_args[0][0]["message"], expected_output)

    @mock.patch("actions.actions.action_fail")
    @mock.patch("actions.actions.action_get")
    @mock.patch("subprocess.check_output")
    def test_subprocess_error(
        self, mock_check_output, mock_action_get, mock_action_fail
    ):
        def raise_error(*args, **kwargs):
            raise subprocess.CalledProcessError(1, ["bogus-cmd"])

        show_all_flag = False
        mock_action_get.return_value = show_all_flag
        mock_check_output.side_effect = raise_error
        show_sel()
        self.assertEqual(
            mock_action_fail.call_args[0][0],
            "Action failed with Command '['bogus-cmd']' "
            "returned non-zero exit status 1.",
        )


class ListCronTestCase(unittest.TestCase):
    @mock.patch("actions.actions.action_set")
    @mock.patch("actions.actions.log")
    @mock.patch("actions.actions.CRON_DIR")
    def test_list_cron_jobs(self, mock_cron_dir, mock_log, mock_action_set):
        mock_cron_dir.glob.return_value = [
            pathlib.Path("/path/to/hwhealth_ipmi"),
            pathlib.Path("/path/to/hwhealth_test"),
        ]

        list_cron_actions()
        mock_cron_dir.glob.assert_called_once_with("hwhealth_*")
        ui_response = "Cron Actions: ipmi,test"
        mock_log.assert_called_once_with(
            "Action list-cron-actions completed\n{}".format(ui_response)
        )
        mock_action_set.assert_called_once_with({"message": ui_response})

    @mock.patch("subprocess.check_output")
    def test_run_good_job_file(self, mock_check_output):
        test_file = pathlib.Path(__file__).parent / "test_run_good_job_file.cron"
        mock_check_output.side_effect = [
            "result #1",
            "result #2",
        ]
        job_output = _run_job_file(test_file)
        assert job_output == [
            (
                [
                    "sudo",
                    "-u",
                    "nagios",
                    "/run/my/command",
                    "with",
                    "some",
                    "--arguments",
                ],
                "result #1",
            ),
            (
                [
                    "sudo",
                    "-u",
                    "root",
                    "/run/my/command",
                    "with",
                    "some",
                    "--other-arguments",
                ],
                "result #2",
            ),
        ]

    @mock.patch("actions.actions.action_fail")
    @mock.patch("subprocess.check_output")
    def test_run_job_file_failure(self, mock_check_output, mock_action_fail):
        with tempfile.TemporaryDirectory() as tempdirname:
            test_file = pathlib.Path(tempdirname) / "test"
            with open(test_file, "w") as fp:
                fp.write("* * * * * root /run/my/command with some --other-arguments")

            mock_check_output.side_effect = [
                subprocess.CalledProcessError(
                    cmd="sudo",
                    returncode=-1,
                    output="No ",
                    stderr="Way",
                ),
            ]
            job_output = _run_job_file(test_file)
            assert job_output == [
                (
                    [
                        "sudo",
                        "-u",
                        "root",
                        "/run/my/command",
                        "with",
                        "some",
                        "--other-arguments",
                    ],
                    "No Way",
                )
            ]
            mock_action_fail.assert_called_with(
                "root: /run/my/command with some --other-arguments\n\t"
                "Command 'sudo' died with <Signals.SIGHUP: 1>."
            )

    @mock.patch("actions.actions._run_job_file")
    @mock.patch("actions.actions._get_cron_actions")
    @mock.patch("actions.actions.action_get")
    @mock.patch("actions.actions.action_set")
    def test_run_cron_action(
        self,
        mock_action_set,
        mock_action_get,
        mock_get_cron_actions,
        mock_run_job_file,
    ):
        mock_action_get.return_value = "test,garbage"
        mock_get_cron_actions.return_value = {"test"}
        mock_run_job_file.return_value = [
            (["sudo", "-u", "root", "/run/my/cmd"], "works!")
        ]
        run_cron_action()
        mock_action_set.assert_called_once_with(
            {
                "message": "sudo -u root /run/my/cmd\n\tworks!",
                "skipped": "unavailable actions: garbage",
            }
        )
        mock_run_job_file.assert_called_once_with(
            pathlib.Path("/etc/cron.d/hwhealth_test")
        )
